@@ -1,5 +1,4 @@
 /* This file is part of GEGL.
- * ck
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -12,7 +11,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, see <http://www.gnu.org/licenses/>.
+ * License along with this library; if not, see <https://www.gnu.org/licenses/>.
  *
  * Copyright 2006-2008 Øyvind Kolås <pippin@gimp.org>
  */
@@ -24,14 +23,13 @@
 #include "gegl-buffer.h"
 #include "gegl-tile-handler.h"
 #include "gegl-buffer-iterator.h"
-#include "gegl-buffer-cl-iterator.h"
-
-// #define GEGL_USE_TILE_MUTEX
 
 #define GEGL_BUFFER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass),  GEGL_TYPE_BUFFER, GeglBufferClass))
 #define GEGL_IS_BUFFER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass),  GEGL_TYPE_BUFFER))
 #define GEGL_BUFFER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj),  GEGL_TYPE_BUFFER, GeglBufferClass))
 
+typedef struct _GeglBufferClass GeglBufferClass;
+
 struct _GeglBuffer
 {
   GeglTileHandler   parent_instance; /* which is a GeglTileHandler which has a
@@ -46,16 +44,13 @@ struct _GeglBuffer
                                 be different from format */
 
   gint              shift_x; /* The relative offset of origins compared with */
-  gint              shift_y; /* anchestral tile_storage buffer, during            */
-                             /* construction relative to immediate source  */
+  gint              shift_y; /* anchestral tile_storage buffer, during       */
+                             /* construction relative to immediate source    */
 
   GeglRectangle     abyss;
-
-
-  GeglSampler      *sampler; /* cached sampler for speeding up random
-                                access interpolated fetches from the
-                                buffer */
-  const Babl       *sampler_format; /* the format of the cached sampler */
+  gboolean          abyss_tracks_extent; /* specifies whether the abyss rectangle
+                                            should track any modifications to the
+                                            extent rectangle */
 
   GeglTileStorage  *tile_storage;
 
@@ -65,10 +60,16 @@ struct _GeglBuffer
 
   gint              lock_count;
 
-  gchar            *alloc_stack_trace; /* Stack trace for allocation,
+  gpointer         *alloc_stack_trace; /* Stack trace for allocation,
                                           useful for debugging */
+  gint              alloc_stack_size;
 
-  gpointer          backend;
+  gint              changed_signal_connections; /* to avoid firing changed signals
+                                                   with no listeners */
+  gint              changed_signal_freeze_count;
+  GeglRectangle     changed_signal_accumulator;
+
+  GeglTileBackend  *backend;
 };
 
 struct _GeglBufferClass
@@ -77,55 +78,60 @@ struct _GeglBufferClass
 };
 
 
-
 gint              gegl_buffer_leaks       (void);
 
 void              gegl_buffer_stats       (void);
 
-void              gegl_buffer_save        (GeglBuffer          *buffer,
-                                           const gchar         *path,
-                                           const GeglRectangle *roi);
-
-
-const gchar      *gegl_swap_dir           (void);
-
-
 void              gegl_tile_cache_init    (void);
 
 void              gegl_tile_cache_destroy (void);
 
+void              gegl_tile_backend_swap_cleanup (void);
+
 GeglTileBackend * gegl_buffer_backend     (GeglBuffer *buffer);
+GeglTileBackend * gegl_buffer_backend2    (GeglBuffer *buffer); /* non-cached */
 
 gboolean          gegl_buffer_is_shared   (GeglBuffer *buffer);
 
+#define GEGL_BUFFER_DISABLE_LOCKS 1
+
+#ifdef GEGL_BUFFER_DISABLE_LOCKS
+#define           gegl_buffer_try_lock(a)   (TRUE)
+#define           gegl_buffer_lock(a)       do{}while(0)
+#define           gegl_buffer_unlock(a)       do{}while(0)
+#else
 gboolean          gegl_buffer_try_lock    (GeglBuffer *buffer);
 gboolean          gegl_buffer_lock        (GeglBuffer *buffer);
 gboolean          gegl_buffer_unlock      (GeglBuffer *buffer);
+#endif
+
 void              gegl_buffer_set_unlocked (GeglBuffer          *buffer,
                                             const GeglRectangle *rect,
+                                            gint                 level,
                                             const Babl          *format,
                                             const void          *src,
                                             gint                 rowstride);
+
+void              gegl_buffer_set_unlocked_no_notify (GeglBuffer          *buffer,
+                                                      const GeglRectangle *rect,
+                                                      gint                 level,
+                                                      const Babl          *format,
+                                                      const void          *src,
+                                                      gint                 rowstride);
+
 void              gegl_buffer_get_unlocked (GeglBuffer          *buffer,
                                             gdouble              scale,
                                             const GeglRectangle *rect,
                                             const Babl          *format,
                                             gpointer             dest_buf,
-                                            gint                 rowstride);
-
-GeglBuffer *
-gegl_buffer_new_ram (const GeglRectangle *extent,
-                     const Babl          *format);
-
-void            gegl_buffer_sampler           (GeglBuffer     *buffer,
-                                               gdouble         x,
-                                               gdouble         y,
-                                               gdouble         scale,
-                                               gpointer        dest,
-                                               const Babl     *format,
-                                               gpointer        sampler);
+                                            gint                 rowstride,
+                                            GeglAbyssPolicy      repeat_mode);
 
+GeglBuffer *      gegl_buffer_new_ram     (const GeglRectangle *extent,
+                                           const Babl          *format);
 
+void              gegl_buffer_emit_changed_signal (GeglBuffer *buffer,
+                                                   const GeglRectangle *rect);
 
 /* the instance size of a GeglTile is a bit large, and should if possible be
  * trimmed down
@@ -150,18 +156,30 @@ struct _GeglTile
   guint            stored_rev;  /* what revision was we when we from tile_storage?
                                    (currently set to 1 when loaded from disk */
 
-  gchar            lock;        /* number of times the tile is write locked
-                                 * should in theory just have the values 0/1
+  gint             lock_count;       /* number of outstanding write locks */
+  gint             read_lock_count;  /* number of outstanding read locks */
+  guint            is_zero_tile:1;   /* whether the tile data is fully zeroed
+                                      * (allowing for false negatives, but not
+                                      * false positives)
+                                      */
+  guint            is_global_tile:1; /* whether the tile data is global (and
+                                      * therefore can never be owned by a
+                                      * single mutable tile)
+                                      */
+  guint            keep_identity:1;  /* maintain data pointer identity, rather
+                                      * than data content only
+                                      */
+
+  gint             clone_state; /* tile clone/unclone state & spinlock */
+  gint            *n_clones;    /* an array of two atomic counters, shared
+                                 * among all tiles sharing the same data.
+                                 * the first counter indicates the number of
+                                 * tiles sharing the data, and the second
+                                 * counter indicates how many of those tiles
+                                 * are in the cache.
                                  */
 
-
-#ifdef GEGL_USE_TILE_MUTEX
-  GMutex          *mutex;
-#endif
-
-  /* the shared list is a doubly linked circular list */
-  GeglTile        *next_shared;
-  GeglTile        *prev_shared;
+  guint64          damage;
 
   /* called when the tile is about to be destroyed */
   GDestroyNotify   destroy_notify;
@@ -174,12 +192,35 @@ struct _GeglTile
   gpointer         unlock_notify_data;
 };
 
+gboolean gegl_tile_needs_store    (GeglTile *tile);
+void     gegl_tile_unlock_no_void (GeglTile *tile);
+gboolean gegl_tile_damage         (GeglTile *tile,
+                                   guint64   damage);
+
 void _gegl_buffer_drop_hot_tile (GeglBuffer *buffer);
 
-#ifndef __GEGL_TILE_C
-#define gegl_tile_get_data(tile)  ((guchar*)((tile)->data))
-#endif // __GEGL_TILE_C
+GeglRectangle _gegl_get_required_for_scale (const GeglRectangle *roi,
+                                            gdouble              scale);
 
+gboolean gegl_buffer_scan_compatible (GeglBuffer *bufferA,
+                                      gint        xA,
+                                      gint        yA,
+                                      GeglBuffer *bufferB,
+                                      gint        xB,
+                                      gint        yB);
+
+
+extern void (*gegl_tile_handler_cache_ext_flush) (void *tile_handler_cache, const GeglRectangle *rect);
+extern void (*gegl_buffer_ext_flush) (GeglBuffer *buffer, const GeglRectangle *rect);
+extern void (*gegl_buffer_ext_invalidate) (GeglBuffer *buffer, const GeglRectangle *rect);
+
+
+#ifndef __GEGL_TILE_H__
+#define gegl_tile_get_data(tile)  ((tile)->data)
+#endif
+
+#define gegl_tile_n_clones(tile)         (&(tile)->n_clones[0])
+#define gegl_tile_n_cached_clones(tile)  (&(tile)->n_clones[1])
 
 /* computes the positive integer remainder (also for negative dividends)
  */
@@ -198,5 +239,4 @@ void _gegl_buffer_drop_hot_tile (GeglBuffer *buffer);
       (coordinate) / (stride):\
       ((((coordinate) + 1) /(stride)) - 1))
 
-
 #endif
